#include <iostream>
#include <vector>
#include <random>
#include <algorithm>

template <typename T>
class Iterator {

public:
	using iterator_category = std::random_access_iterator_tag;
	using value_type      = T;
	using reference_type  = T&;
	using pointer_type    = T*;
	using difference_type = ptrdiff_t;

	// All iterators must be constructible, copy-constructible, copy-assignable,
	// destructible and swappable.

	// The custom constructor satisfies the constructible requirement,
	// while all others are covered by the implicitly-declared constructors
	// and operators kindly provided by the compiler.

	explicit Iterator(pointer_type ptr = nullptr) : ptr(ptr) {}

	// dereferenceable
	reference_type operator* () const { return *ptr; }
	pointer_type   operator->() const { return  ptr; }

	// incrementable / decrementable
	Iterator& operator++()    { ptr++; return *this; }
	Iterator  operator++(int) { auto tmp = *this; ++(*this); return tmp; }
	Iterator& operator--()    { ptr--; return *this; }
	Iterator  operator--(int) { auto tmp = *this; --(*this); return tmp; }

	difference_type operator-(const Iterator& it)   const { return this->ptr - it.ptr; }
	Iterator operator+(const difference_type& diff) const { return Iterator(ptr + diff); }
	Iterator operator-(const difference_type& diff) const { return Iterator(ptr - diff); }

	reference_type operator[] (const difference_type& offset) const { return *(*this + offset); }

	// comparable
	auto operator<=> (const Iterator& it) const = default;

private:
	pointer_type ptr;
};

template <typename T>
class Y {

public:
	using iterator = Iterator<T>;
	using const_iterator = Iterator<const T>;
	using reverse_iterator = std::reverse_iterator<iterator>;
	using const_reverse_iterator = std::reverse_iterator<const_iterator>;

	iterator begin() { return iterator(&array[0]); }
	iterator end() { return iterator(&array[N]); }
	[[nodiscard]]
	size_t size() const noexcept { return N; }

	[[nodiscard]]
	const_iterator cbegin() const { return const_iterator(&array[0]); }
	[[nodiscard]]
	const_iterator cend() const { return const_iterator(&array[N]); }

	reverse_iterator rbegin() { return reverse_iterator(end()); }
	reverse_iterator rend() { return reverse_iterator(begin()); }

	[[nodiscard]]
	const_reverse_iterator crbegin() const {
		return const_reverse_iterator(cend());
	}
	[[nodiscard]]
	const_reverse_iterator crend() const {
		return const_reverse_iterator(cbegin());
	}

//	Y() : array(N) { }

private:
	static constexpr int N{16};
	T array[N];
//	std::vector<T> array{};
};

void print_range(Y<int> y) {
	std::ranges::for_each(y, [] (int i) {
		std::cout << i << " ";
	});
	std::cout << std::endl;
}

// The above is translated by the compiler into:
void print_iterator(const Y<int>& y) {
	for (auto it = y.cbegin(); it != y.cend(); ++it) {
		std::cout << *it << ' ';
	}
	std::cout << std::endl;
}

void print_reverse(const Y<int>& y) {
	for (auto it = y.crbegin(); it != y.crend(); ++it) {
		std::cout << *it << ' ';
	}
	std::cout << std::endl;
}

int main() {
	std::random_device dev;
	std::mt19937 rng(dev());
	// uniform distribution in [0, 99]
	std::uniform_int_distribution<std::mt19937::result_type> dist(0,99);

	Y<int> y{};
	for (auto it = y.begin(); it != y.end(); ++it) {
		*it = (int)dist(rng);
	}
	print_iterator(y);
	print_range(y);

	std::sort(y.begin(), y.end(), std::less<>());
	print_iterator(y);
	print_reverse(y);

	std::shuffle(y.begin(), y.end(), rng);
	print_iterator(y);

	auto cnt = std::ranges::count_if(y,
		 [] (int i) { return i > 50; });
	std::cout << "Count: " << cnt << std::endl;
}

